"""
Otter Grade command-line utility. Provides local grading of submissions in parallel Docker containers.
"""

import re
import os
import pandas as pd

from .metadata import GradescopeParser, CanvasParser, JSONParser, YAMLParser
from .containers import launch_grade
from .utils import merge_csv, prune_images

def main(path, output_dir, autograder, gradescope, canvas, json, yaml, containers, scripts, no_kill, 
        debug, zips, image, pdfs, prune, force, verbose, **kwargs):
    """
    Runs Otter Grade

    Grades a directory of submissions in parallel Docker containers. Results are outputted as a CSV file
    called ``final_grades.csv``. If ``prune`` is ``True``, Otter's dangling grading images are pruned 
    and the program exits.

    Args:
        path (``str``): path to directory of submissions
        output_dir (``str``): directory in which to write ``final_grades.csv``
        autograder (``str``): path to Otter autograder configuration zip file
        gradescope (``bool``): whether submissions are a Gradescope export
        canvas (``bool``): whether submissions are a Canvas export
        json (``str``): path to a JSON metadata file
        yaml (``str``): path to a YAML metadata file
        containers (``int``): number of containers to run in parallel
        scripts (``bool``): whether Python scripts are being graded
        no_kill (``bool``): whether to keep containers after grading is finished
        debug (``bool``): whether to print the stdout of each container
        zips (``bool``): whether the submissions are Otter-exported zip files
        image (``bool``): base image from which to build grading image
        pdfs (``bool``): whether to copy notebook PDFs out of the containers
        prune (``bool``): whether to prune dangling grading images
        force (``bool``): whether to force pruning (without confirmation)
        verbose (``bool``): whether to log status messages to stdout
        **kwargs: ignored kwargs (a remnant of how the argument parser is built)

    Raises:
        ``AssertionError``: if invalid arguments are provided
    """
    # prune images
    if prune:
        if not force:
            sure = input("Are you sure you want to prune Otter's grading images? This action cannot be undone [y/N] ")
            sure = bool(re.match(sure, r"ye?s?", flags=re.IGNORECASE))
        else:
            sure = True
        
        if sure:
            prune_images()
        
        return

    # Asserts that exactly one metadata flag is provided
    assert sum([meta != False for meta in [
        gradescope,
        canvas,
        json,
        yaml
    ]]) <= 1, "You can specify at most one metadata flag (-g, -j, -y, -c)"

    # verbose flag
    verbose = verbose

    # Hand off metadata to parser
    if gradescope:
        meta_parser = GradescopeParser(path)
        if verbose:
            print("Found Gradescope metadata...")
    elif canvas:
        meta_parser = CanvasParser(path)
        if verbose:
            print("Found Canvas metadata...")
    elif json:
        meta_parser = JSONParser(os.path.join(json))
        if verbose:
            print("Found JSON metadata...")
    elif yaml:
        meta_parser = YAMLParser(os.path.join(yaml))
        if verbose:
            print("Found YAML metadata...")
    else:
        meta_parser = None

    if verbose:
        print("Launching docker containers...")

    #Docker
    grade_dfs = launch_grade(autograder,
        notebooks_dir=path,
        verbose=verbose,
        num_containers=containers,
        scripts=scripts,
        no_kill=no_kill,
        output_path=output_dir,
        debug=debug,
        zips=zips,
        image=image,
        meta_parser=meta_parser,
        pdfs=pdfs
    )

    if verbose:
        print("Combining grades and saving...")

    # Merge Dataframes
    output_df = merge_csv(grade_dfs)

    def map_files_to_ids(row):
        """Returns the identifier for the filename in the specified row"""
        return meta_parser.file_to_id(row["file"])

    # add in identifier column
    if meta_parser is not None:
        output_df["identifier"] = output_df.apply(map_files_to_ids, axis=1)
        output_df.drop("file", axis=1, inplace=True)

        # reorder cols in output_df
        cols = output_df.columns.tolist()
        output_df = output_df[cols[-1:] + cols[:-1]]

    # write to CSV file
    output_df.to_csv(os.path.join(output_dir, "final_grades.csv"), index=False)



###############################
##### Tests for otter run #####
###############################

import os
import unittest
import subprocess
import json
import shutil
import nbformat
import nbconvert

from subprocess import PIPE
from glob import glob
from unittest import mock
from shutil import copyfile

# from otter.argparser import get_parser
from otter.generate.autograder import main as autograder
from otter.run.run_autograder import main as run_autograder

from . import TestCase

# parser = get_parser()

NBFORMAT_VERSION = 4
TEST_FILES_PATH = "test/test-run/"

class TestRun(TestCase):
    def setUp(self):
        super().setUp()

        self.cwd = os.getcwd()

        with open(TEST_FILES_PATH + "autograder/source/otter_config.json") as f:
            self.config = json.load(f)
        
        self.expected_results = {
            "tests": [
                {
                    "name": "q1 - 1",
                    "score": 0.0,
                    "max_score": 0.0,
                    "visibility": "visible",
                    "output": "Test case passed!"
                },
                {
                    "name": "q1 - 2",
                    "score": 0.0,
                    "max_score": 0.0,
                    "visibility": "visible",
                    "output": "Test case passed!"
                },
                {
                    "name": "q1 - 3",
                    "score": 0.0,
                    "max_score": 0.0,
                    "visibility": "hidden",
                    "output": "Test case passed!"
                },
                {
                    "name": "q1 - 4",
                    "score": 0.0,
                    "max_score": 0.0,
                    "visibility": "hidden",
                    "output": "Test case passed!"
                },
                {
                    "name": "q2 - 1",
                    "score": 0.0,
                    "max_score": 0.5,
                    "visibility": "visible",
                    "output": "Trying:\n    negate(True)\nExpecting:\n    False\n**********************************************************************\nLine 2, in q2 0\nFailed example:\n    negate(True)\nExpected:\n    False\nGot:\n    True\n"
                },
                {
                    "name": "q2 - 2",
                    "score": 0.0,
                    "max_score": 0.5,
                    "visibility": "visible",
                    "output": "Trying:\n    negate(False)\nExpecting:\n    True\n**********************************************************************\nLine 2, in q2 1\nFailed example:\n    negate(False)\nExpected:\n    True\nGot:\n    False\n"
                },
                {
                    "name": "q2 - 3",
                    "score": 0.0,
                    "max_score": 0.5,
                    "visibility": "hidden",
                    "output": "Trying:\n    negate(\"\")\nExpecting:\n    True\n**********************************************************************\nLine 2, in q2 2\nFailed example:\n    negate(\"\")\nExpected:\n    True\nGot:\n    ''\n"
                },
                {
                    "name": "q2 - 4",
                    "score": 0.0,
                    "max_score": 0.5,
                    "visibility": "hidden",
                    "output": "Trying:\n    negate(1)\nExpecting:\n    False\n**********************************************************************\nLine 2, in q2 3\nFailed example:\n    negate(1)\nExpected:\n    False\nGot:\n    1\n"
                },
                {
                    "name": "q3 - 1",
                    "score": 1.0,
                    "max_score": 1.0,
                    "visibility": "visible",
                    "output": "Test case passed!"
                },
                {
                    "name": "q3 - 2",
                    "score": 1.0,
                    "max_score": 1.0,
                    "visibility": "hidden",
                    "output": "Test case passed!"
                },
                {
                    "name": "q4 - 1",
                    "score": 1.0,
                    "max_score": 1.0,
                    "visibility": "hidden",
                    "output": "Test case passed!"
                },
                {
                    "name": "q6 - 1",
                    "score": 2.5,
                    "max_score": 2.5,
                    "visibility": "visible",
                    "output": "Test case passed!"
                },
                {
                    "name": "q6 - 2",
                    "score": 0.0,
                    "max_score": 2.5,
                    "visibility": "hidden",
                    "output": "Trying:\n    fib = fiberator()\nExpecting nothing\nok\nTrying:\n    for _ in range(10):\n        print(next(fib))\nExpecting:\n    0\n    1\n    1\n    2\n    3\n    5\n    8\n    13\n    21\n    34\n**********************************************************************\nLine 3, in q6 1\nFailed example:\n    for _ in range(10):\n        print(next(fib))\nExpected:\n    0\n    1\n    1\n    2\n    3\n    5\n    8\n    13\n    21\n    34\nGot:\n    0\n    1\n    1\n    1\n    2\n    3\n    5\n    8\n    13\n    21\n"
                },
                {
                    "name": "q7 - 1",
                    "score": 1.0,
                    "max_score": 1.0,
                    "visibility": "visible",
                    "output": "Test case passed!"
                }
            ],
            "output": "Students are allowed 1 submissions every 1 days. You have 0 submissions in that period."
        }

    def test_notebook(self):

        run_autograder(self.config['autograder_dir'])

        with open(TEST_FILES_PATH + "autograder/results/results.json") as f:
            actual_results = json.load(f)

        self.assertEqual(actual_results, self.expected_results, f"Actual results did not matched expected:\n{actual_results}")

    def test_script(self):

        nb_path = TEST_FILES_PATH + "autograder/submission/fails2and6H.ipynb"
        nb = nbformat.read(nb_path, as_version=NBFORMAT_VERSION)
        os.remove(nb_path)

        try:
            py, _ = nbconvert.export(nbconvert.PythonExporter, nb)
            
            # remove magic commands
            py = "\n".join(l for l in py.split("\n") if not l.startswith("get_ipython"))
            
            with open(TEST_FILES_PATH + "autograder/submission/fails2and6H.py", "w+") as f:
                f.write(py)

            run_autograder(self.config['autograder_dir'])

            with open(TEST_FILES_PATH + "autograder/results/results.json") as f:
                actual_results = json.load(f)

            self.assertEqual(actual_results, self.expected_results, f"Actual results did not matched expected:\n{actual_results}")

            self.deletePaths([TEST_FILES_PATH + "autograder/submission/fails2and6H.py"])

            with open(nb_path, "w+") as f:
                nbformat.write(nb, f)
        
        except:
            os.chdir(self.cwd)
            self.deletePaths([TEST_FILES_PATH + "autograder/submission/fails2and6H.py"])
            with open(nb_path, "w+") as f:
                nbformat.write(nb, f)
            
            raise

    def tearDown(self):
        os.chdir(self.cwd)
        self.deletePaths([
            TEST_FILES_PATH + "autograder/results/results.json",
            TEST_FILES_PATH + "autograder/results/results.pkl",
            TEST_FILES_PATH + "autograder/__init__.py",
            TEST_FILES_PATH + "autograder/submission/test",
            TEST_FILES_PATH + "autograder/submission/tests",
            TEST_FILES_PATH + "autograder/submission/__init__.py",
            TEST_FILES_PATH + "autograder/submission/.OTTER_LOG",
        ])



"""
Docker container management for Otter Grade
"""

import json
import subprocess
import re
import os
import shutil
import tempfile
import docker
import pandas as pd
import tarfile
import pkg_resources
import glob
import pickle

from subprocess import PIPE
from concurrent.futures import ThreadPoolExecutor, wait

from .metadata import GradescopeParser
from .utils import simple_tar, get_container_file, generate_hash, OTTER_DOCKER_IMAGE_TAG

from ..test_files import GradingResults


def build_image(zip_path, base_image, tag):
    """
    Creates a grading image based on the autograder zip file and attaches a tag.

    Args:
        zip_path (``str``): path to the autograder zip file
        base_image (``str``): base Docker image to build from
        tag (``str``): tag to be added when creating the image

    Returns:
        ``str``: the tag of the newly-build Docker image
    """
    image = OTTER_DOCKER_IMAGE_TAG + ":" + tag
    dockerfile = pkg_resources.resource_filename(__name__, "Dockerfile")
    build_out = subprocess.Popen(
        ["docker", "build","--build-arg", "ZIPPATH=" + zip_path, "--build-arg", "BASE_IMAGE=" + base_image,
         ".", "-f", dockerfile, "-t", image],
    )
    build_out.wait()
    return image

def launch_grade(zip_path, notebooks_dir, verbose=False, num_containers=None,
    scripts=False, no_kill=False, output_path="./", debug=False, zips=False, image="ucbdsinfra/otter-grader",
    meta_parser=None, pdfs=False):
    """
    Grades notebooks in parallel Docker containers

    This function runs ``num_containers`` Docker containers in parallel to grade the student submissions
    in ``notebooks_dir`` using the autograder configuration file at ``zip_path``. It can additionally 
    generate PDFs for the parts of the assignment needing manual grading.

    Args:
        zip_path(``str``): path to zip file used to set up container
        notebooks_dir (``str``): path to directory of student submissions to be graded
        verbose (``bool``, optional): whether status messages should be printed to the command line
        num_containers (``int``, optional): The number of parallel containers that will be run
        scripts (``bool``, optional): whether student submissions are Python scripts rather than
            Jupyter notebooks
        no_kill (``bool``, optional): whether the grading containers should be kept running after
            grading finishes
        output_path (``str``, optional): path at which to write grades CSVs copied from the container
        debug (``bool``, optional): whether to run grading in debug mode (prints grading STDOUT and STDERR
            from each container to the command line)
        zips (``bool``, optional): whether the submissions are zip files formatted from ``Notebook.export``
        image (``str``, optional): a base image to use for building Docker images
        meta_parser (object, optional): a metadata parser instance; one of the classes defined in
            ``otter.metadata``
        pdfs (``bool``, optional): whether to copy PDFs out of the containers

    Returns:
        ``list`` of ``pandas.core.frame.DataFrame``: the grades returned by each container spawned during
            grading
    """
    if not num_containers:
        num_containers = 4

    pool = ThreadPoolExecutor(num_containers)
    futures = []
    img = build_image(zip_path, image, generate_hash(zip_path))

    notebooks = glob.glob(os.path.join(notebooks_dir, "*.ipynb"))
    for nb_path in notebooks:
        futures += [pool.submit(grade_assignments,
            notebook_dir=nb_path,
            verbose=verbose,
            #TODO:check if path is not default for generate hash
            image=img,
            scripts=scripts,
            no_kill=no_kill,
            output_path=output_path,
            debug=debug,
            zips=zips,
            pdfs=pdfs
        )]

    # stop execution while containers are running
    finished_futures = wait(futures)

    # return list of dataframes
    return [df.result() for df in finished_futures[0]]


# TODO: these arguments need to be updated. replace notebooks_dir with the path to the notebook that
# this container will be grading
def grade_assignments(notebook_dir, image="ucbdsinfra/otter-grader", verbose=False,
    scripts=False, no_kill=False, output_path="./", debug=False, zips=False, pdfs=False):
    """
    Grades multiple submissions in a directory using a single docker container. If no PDF assignment is
    wanted, set all three PDF params (``unfiltered_pdfs``, ``tag_filter``, and ``html_filter``) to ``False``.

    Args:
        notebook_dir (``str``): path to directory of student submissions to be graded
        image (``str``, optional): a Docker image tag to be used for grading environment
        verbose (``bool``, optional): whether status messages should be printed to the command line
        scripts (``bool``, optional): whether student submissions are Python scripts rather than
            Jupyter notebooks
        no_kill (``bool``, optional): whether the grading containers should be kept running after
            grading finishes
        output_path (``str``, optional): path at which to write grades CSVs copied from the container
        debug (``bool``, False): whether to run grading in debug mode (prints grading STDOUT and STDERR
            from each container to the command line)
        zips (``bool``, False): whether the submissions are zip files formatted from ``Notebook.export``
        pdfs (``bool``, optional): whether to copy PDFs out of the containers

    Returns:
        ``pandas.core.frame.DataFrame``: A dataframe of file to grades information
    """
    # this is a fix for travis -- allows overriding docker client version
    if os.environ.get("OTTER_DOCKER_CLIENT_VERSION") is not None:
        client = docker.from_env(version=os.environ.get("OTTER_DOCKER_CLIENT_VERSION"))
    else:
        client = docker.from_env()
    container = client.containers.run(image, detach=True, tty=True)

    notebook_dir = os.path.abspath(notebook_dir)
    nb_name = os.path.splitext(os.path.split(notebook_dir)[1])[0]

    try:
        container_id = container.id[:12]

        if verbose:
            print(f"Launched container {container_id}...")

        # TODO: remember 1 subm per container, so we will use this content manager to put the
        #       notebook at /autograder/submission/{notebook name}.ipynb

        with simple_tar(notebook_dir) as tarf:
            container.put_archive("/autograder/submission", tarf)

        if verbose:
            print(f"Grading {('notebooks', 'scripts')[scripts]} in container {container_id}...")

        # Now we have the notebooks in home/notebooks, we should tell the container to execute the grade command...
        grade_command = ["/autograder/run_autograder"]

        exit_code, output = container.exec_run(grade_command)
        assert exit_code == 0, f"Container {container_id} failed with output:\n{output.decode('utf-8')}"

        if debug:
            print(output.decode("utf-8"))

        if verbose:
            print(f"Copying grades from container {container_id}...")

        # TODO: this needs to be updated to parse the JSON, found at path /autograder/results/results.json,
        #       into the existing CSV format so that this function returns a 1-row dataframe
        # get the grades back from the container and read to date frame so we can merge later
        
        # with get_container_file(container, "/home/notebooks/grades.csv") as f:
        #     df = pd.read_csv(f)
        
        #should be fixed @Edward
        with get_container_file(container, "/autograder/results/results.pkl") as f:
            scores = pickle.load(f)

        # TODO: wrangle results
        scores = scores.to_dict()
        scores = {t : [scores[t]["score"]] if type(scores[t]) == dict else scores[t] for t in scores}
        scores["file"] = os.path.split(notebook_dir)[1]
        df = pd.DataFrame(scores)

        # TODO: PDFs still need to work, so this code needs to be adapted to get the PDF of the notebook
        #       at path /autograder/submission/{notebook name}.pdf

        #not fixed yet @Edward

        if pdfs:
            with get_container_file(container, f"/autograder/submission/{nb_name}.pdf") as pdf_file:
                pdf_folder = os.path.join(os.path.abspath(output_path), "submission_pdfs")
                os.makedirs(pdf_folder, exist_ok=True)
            
                # # copy out manual submissions
                # for pdf in df["manual"]:
                local_pdf_path = os.path.join(pdf_folder, f"{nb_name}.pdf")
                with open(local_pdf_path, "wb+") as f:
                    f.write(pdf_file.read())
        
            # df["manual"] = df["manual"].str.replace("/home/notebooks", os.path.basename(pdf_folder))

        if not no_kill:
            if verbose:
                print(f"Stopping container {container_id}...")

            # cleanup the docker container
            container.stop()
            container.remove()

        client.close()

    except:

        # # delete the submission PDFs on failure
        # if pdfs and "pdf_folder" in locals():
        #     shutil.rmtree(pdf_folder)

        if not no_kill:
            if verbose:
                print(f"Stopping container {container_id}...")

            # cleanup the docker container
            container.stop()
            container.remove()

        client.close()

        raise

    return df



"""
Abstract base classes for working with test files and classes to represent collections of test and 
their results
"""

from abc import ABC, abstractmethod
from collections import namedtuple
from textwrap import dedent
from typing import Tuple, List, Dict, Any
from jinja2 import Template
from pygments import highlight
from pygments.lexers import PythonConsoleLexer
from pygments.formatters import HtmlFormatter


# class for storing the test cases themselves
#   - body is the string that gets run for the test
#   - hidden is the visibility of the test case
TestCase = namedtuple("TestCase", ["name", "body", "hidden"])


# class for storing the results of a single test _case_ (within a test file)
#   - message should be a string to print out to the student (ignored if passed is True)
#   - passed is whether the test case passed
#   - hidden is the visibility of the test case
TestCaseResult = namedtuple("TestCaseResult", ["test_case", "message", "passed"])


# TODO: fix reprs
class TestFile(ABC):
    """
    A (abstract) single test file for Otter. This ABC defines how test results are represented and sets
    up the instance variables tracked by tests. Subclasses should implement the abstract class method
    ``from_file`` and the abstract instance method ``run``.

    Args:
        name (``str``): the name of test file
        path (``str``): the path to the test file
        test_cases (``list`` of ``TestCase``): a list of parsed tests to be run
        value (``int``, optional): the point value of this test, defaults to 1
        all_or_nothing (``bool``, optional): whether the test should be graded all-or-nothing across
            cases

    Attributes:
        name (``str``): the name of test file
        path (``str``): the path to the test file
        test_cases (``list`` of ``TestCase``): a list of parsed tests to be run
        value (``int``): the point value of this test, defaults to 1
        all_or_nothing (``bool``): whether the test should be graded all-or-nothing across
            cases
        passed_all (``bool``): whether all of the test cases were passed
        test_case_results (``list`` of ``TestCaseResult``): a list of results for the test cases in
            ``test_cases``
        grade (``float``): the percentage of ``points`` earned for this test file as a decimal
    """

    html_result_pass_template = Template("""
    <p><strong>{{ name }}</strong> passed!</p>
    """)

    plain_result_pass_template = Template("{{ name }} passed!")

    html_result_fail_template = Template("""
    <p><strong style='color: red;'>{{ name }}</strong></p>
    <p><strong>Test result:</strong></p>
    {% for test_case_result in test_case_results %}
        <p><em>{{ test_case_result.test_case.name }}</em>
        {% if not test_case_result.passed %}
            <pre>{{ test_case_result.message }}</pre>
        {% else %}
            <pre>Test case passed!</pre>
        {% endif %}
        </p>
    {% endfor %}
    """)

    plain_result_fail_template = Template(dedent("""\
    {{ name }} results:
    {% for test_case_result in test_case_results %}{% if not test_case_result.passed %}
    {{ test_case_result.message }}{% endif %}{% endfor %}"""))

    def _repr_html_(self):
        if self.passed_all:
            return type(self).html_result_pass_template.render(name=self.name)
        else:
            return type(self).html_result_fail_template.render(
                name=self.name,
                # test_code=highlight(self.failed_test, PythonConsoleLexer(), HtmlFormatter(noclasses=True)),
                test_case_results=self.test_case_results
            )

    def __repr__(self):
        if self.passed_all:
            return type(self).plain_result_pass_template.render(name=self.name)
        else:
            return type(self).plain_result_fail_template.render(
                name=self.name,
                # test_code=self.failed_test,
                test_case_results=self.test_case_results
            )

    # @abstractmethod
    def __init__(self, name, path, test_cases, value=1, all_or_nothing=True):
        self.name = name
        self.path = path
        # self.public_tests = [t for t, h in zip(tests, hiddens) if not h]
        # self.hidden_tests = [t for t, h in zip(tests, hiddens) if h]
        self.test_cases = test_cases
        self.value = value
        # self.hidden = hidden
        self.passed_all = None
        # self.failed_test = None
        # self.failed_test_hidden = None
        # self.result = None
        self.all_or_nothing = all_or_nothing
        self.test_case_results = []
        self.grade = None

    @classmethod
    @abstractmethod
    def from_file(cls, path):
        ...

    @abstractmethod
    def run(self, global_environment):
        ...



"""
Grading plugins for Otter
"""

import importlib

from .abstract_plugin import AbstractOtterPlugin, PluginEventNotSupportedException


class PluginCollection:
    """
    Class for loading, organizing, and running plugins during grading. This class is instantiated with
    a list of plugin names, which should be importable strings that evaluate to objects that inherit 
    from ``otter.plugins.AbstractOtterPlugin``.

    When this class is instantiated, each plugin is imported and passed its configurations specified
    in the ``otter_config.json``. Plugins should be listed in ``otter_config.json`` in the ``plugins``
    key:

    .. code-block:: json

        {
            "plugins": [
                "some_otter_plugin_package.SomeOtterPlugin",
                {
                    "some_other_otter_plugin_package.SomeOtherOtterPlugin": {
                        "some_key": "some_value"
                    }
                }
            ]
        }

    Args:
        plugin_names (``list[Union[str,dict[str:Any]]]``): the importable names of plugin classes (e.g. 
            ``some_package.SomePlugin``) and their configurations
        submission_path (``str``): the absolute path to the submission being graded
        submission_metadata (``dict[str:Any]``): submission metadata
    """

    @staticmethod
    def _parse_plugin_config(plugin_config):
        if not isinstance(plugin_config, list):
            raise ValueError(f"Invalid plugin config: {plugin_config}")
        
        result = []
        for plg in plugin_config:
            if isinstance(plg, str):
                result.append({
                    "plugin": plg,
                    "config": {},
                })
            elif isinstance(plg, dict):
                keys = list(plg.keys())
                if not len(keys) == 1:
                    raise ValueError(f"Invalid plugin specification: {plg}")
                result.append({
                    "plugin": keys[0],
                    "config": plg[keys[0]],
                })

        return result

    def __init__(self, plugins, submission_path, submission_metadata):
        self._plugin_config = self._parse_plugin_config(plugins)
        self._plugins = None

        self._load_plugins(submission_path, submission_metadata)

    @property
    def _plugin_names(self):
        """
        The importable names of all of the plugins tracked
        """
        return [p["plugin"] for p in self._plugin_config]

    def _load_plugins(self, submission_path, submission_metadata):
        """
        Loads each plugin in ``self._plugin_config`` by importing it with ``importlib`` and creating
        and instance with the ``submission_metadata`` and the configurations from ``self._plugin_config``
        for that plugin. Sets ``self._plugins`` to be the list of imported and instantiated plugins.

        Args:
            submission_path (``str``): the absolute path to the submission being graded
            submission_metadata (``dict``): submission metadata
        """
        plugins = []
        for plg_cfg in self._plugin_config:
            plg, cfg = plg_cfg["plugin"], plg_cfg["config"]
            module, class_ = ".".join(plg.split(".")[:-1]), plg.split(".")[-1]
            module = importlib.import_module(module)
            class_ = getattr(module, class_)

            # get the config key for the plugin
            # plugin_cfg = plugin_config.get(class_.PLUGIN_CONFIG_KEY, {})
            plugin = class_(submission_path, submission_metadata, cfg)
            plugins.append(plugin)

        self._plugins = plugins

    def run(self, event, *args, **kwargs):
        """
        Runs the method ``event`` of each plugin in this collection. Passes ``args`` and ``kwargs``
        to this method. Ignores plugins that raise ``PluginEventNotSupportedException`` for this event.

        Args:
            event (``str``): name of the method of the plugin to run
            *args, **kwargs (any): arguments for the method
        
        Returns:
            ``list[Any]``: the values returned by each plugin for the called event
        """
        # TODO: logging to stdout
        rets = []
        for plugin in self._plugins:
            try:
                if hasattr(plugin, event):
                    ret = getattr(plugin, event)(*args, **kwargs)
                    rets.append(ret)
                else:
                    rets.append(None)
            except PluginEventNotSupportedException:
                rets.append(None)
        return rets

    def before_execution(self, submission):
        """
        Runs the ``before_execution`` event for each plugin, composing the results of each (i.e. the
        transformed notebook returned by one plugin is passed to the next plugin).

        Args:
            submission (``Union[str,nbformat.NotebookNode]``): the submission to be executed
        """
        event = "before_execution"
        for plugin in self._plugins:
            try:
                if hasattr(plugin, event):
                    submission = getattr(plugin, event)(submission)
            except PluginEventNotSupportedException:
                pass
        return submission

    def generate_report(self):
        """
        Runs the ``generate_report`` event of each plugin, formatting and concatenating them into a
        single string and returning it.

        Returns:
            ``str``: the plugin report
        """
        reports = self.run("generate_report")

        header = "=" * 35 + " PLUGIN REPORT " + "=" * 35
        footer = "=" * len(header)

        report = header
        for r, plg in zip(reports, self._plugin_names):
            title = f" {plg} Report "
            dashes = len(header) - len(title)
            if dashes > 4:
                if dashes % 2 == 0:
                    ld, rd = dashes // 2, dashes // 2
                else:
                    ld, rd = dashes // 2, dashes // 2 + 1
            else:
                ld, rd = 2, 2

            title = "-" * ld + title + "-" * rd

            body = title + "\n" + r + "\n"

            report += "\n" + body
        
        report += "\n" + footer

        return report
